/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __DS_ASSISTANT_H__
#define __DS_ASSISTANT_H__

#include <digitalSTROM/dsm-api-v2/dsm-api.h>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <boost/thread/mutex.hpp>
#include <pthread.h>
#include <list>
#include <string>

#include "dsmapi-wrapper.h"
#include "tasks.h"

namespace dSA
{

class dS_Assistant : public dSMApiListener
{
public:
    static dS_Assistant *createInstance();
    static dS_Assistant *getInstance();
    virtual ~dS_Assistant();

    boost::shared_ptr<dSMApiWrapper> dS485();

    void run();
    void shutdown();
    void reopenLog();
    void loadDatabaseDefaults();

    void clearPendingTasks();
    void cancelTask(long long int task_id);
    void pauseTask(long long int task_id);
    void addTask(boost::shared_ptr<dSTask> task);
    void scheduleTask(boost::shared_ptr<dSTask> task, unsigned int seconds);
    void scheduleLongTask(boost::shared_ptr<dSLongTask> task,
                          unsigned int seconds);
    long long int addLongTask(boost::shared_ptr<dSLongTask> task);
    long long int getInvalidTaskID() const;
    boost::shared_ptr<const dSLongTaskInfo> getCurrentTaskInfo() const;
    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
                                                        getTaskList() const;

    bus_state_t getBusState() const;

    void scan(bool async = true);

    const dsid_t getDSID() const;

    virtual void busStateCallback(bus_state_t state);
    virtual void busChangeCallback(dsid_t *id, int flag);
    virtual void deviceAddedCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, uint32_t serial_number);
    virtual void deviceRemovedCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, uint32_t serial_number);
    virtual void deviceStatisticsCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t offset, uint16_t value);
    virtual void deviceActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr, uint8_t click_type);
    virtual void deviceLocalActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t state);
    virtual void deviceConfigCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value);
    virtual void deviceTestTransmissionQualityCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality);
    virtual void commUpstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type,
            uint8_t quality, uint8_t flags, uint8_t cflags);
    virtual void commUpstreamLongCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier, uint32_t data,
            uint8_t quality, uint8_t flags, uint8_t cflags);
    virtual void commDownstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);
    virtual void commDownstreamLong0Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);
    virtual void commDownstreamLong8Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint8_t data);
    virtual void commDownstreamLong16Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command, uint16_t data);

private:
    dS_Assistant();

    static dS_Assistant* m_instance;

    std::string m_logfile_name;
    FILE* m_logfile;
    FILE* m_logfile_traffic;

    bool m_shutdown_flag;
    bool m_running;
    boost::shared_ptr<dSMApiWrapper> m_dsm_api_wrapper;
    boost::shared_ptr<BasicMonitorTask> m_comm_monitor_task;
    boost::shared_ptr<DatabaseCleanupTask> m_db_cleanup_task;

    bus_state_t m_bus_state;
    dsid_t m_own_dsid;

    mutable boost::mutex m_dsa_mutex;
    pthread_t m_task_thread;
    pthread_t m_long_task_thread;
    pthread_mutex_t m_task_list_mutex;
    mutable pthread_mutex_t m_long_task_list_mutex;
    std::list<boost::shared_ptr<dSTask> > m_task_list;
    std::list<boost::shared_ptr<dSLongTask> > m_long_task_list;
    std::list<boost::shared_ptr<dSLongTask> > m_high_prio_task_list;
    std::list<boost::shared_ptr<dSLongTask> > m_paused_task_list;
    std::list<boost::weak_ptr<LongTaskScheduler> > m_scheduled_task_list;
    pthread_cond_t m_wakeup_task_cond;
    pthread_cond_t m_wakeup_long_task_cond;
    boost::shared_ptr<dSLongTaskInfo> m_current_task_info;

    void taskThread();
    void longTaskThread();
    static void *staticThreadProc(void *arg);
    static void *staticLongTaskThreadProc(void *arg);

    void lockTaskList();
    void unlockTaskList();

    void lockLongTaskList() const;
    void unlockLongTaskList() const;

    void connect();
    void disconnect();
};

}; //namespace
#endif
