/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include "json_dsm_web_handler.h"
#include "ds-assistant.h"
#include "exceptions.h"
#include "storage.h"
#include "database.h"
#include "hwstats.h"
#include "log.h"

namespace dSA
{

JSONdSMWebHandler::JSONdSMWebHandler(struct mg_connection* connection,
                               const struct mg_request_info* request_info) :
                               JSONWebHandler(connection, request_info) {}

boost::shared_ptr<const dSM> JSONdSMWebHandler::getDSMfromParam()
{
    std::string dsid = m_parameters["dsid"];
    if (dsid.empty())
    {
        throw _Exception("missing dSM dSID parameter");
    }

    dsid_t dsm_dsid = util::str2dsid(dsid);
    boost::shared_ptr<dSM> dsm = Storage::getInstance()->getDSM(dsm_dsid);
    return dsm;
}

std::string JSONdSMWebHandler::restartRegistration()
{
    std::ostringstream result;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    dS_Assistant::getInstance()->dS485()->restartRegistration(dsm->getDSID());

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdSMWebHandler::getCurrentTemperature()
{
    std::ostringstream result;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    uint16_t t;
    dS_Assistant::getInstance()->dS485()->getDSMtemperature(dsm->getDSID(), &t);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "temperature", json_object_new_int(t));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdSMWebHandler::blink()
{
    std::ostringstream result;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    uint8_t flags;
    dS_Assistant::getInstance()->dS485()->getDisplayMode(dsm->getDSID(), &flags);
    dS_Assistant::getInstance()->dS485()->setDisplayMode(dsm->getDSID(), 0x41);
    sleep(4);
    dS_Assistant::getInstance()->dS485()->setDisplayMode(dsm->getDSID(), flags);

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdSMWebHandler::getStatistic()
{
    std::ostringstream result;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    uint8_t length;
    uint8_t index;
    uint8_t data[256];
    try
    {
        index = 0;
        dS_Assistant::getInstance()->dS485()->getStatisticUpstream(
                                    dsm->getDSID(), 0, &length, data + index);
        index += length;
        dS_Assistant::getInstance()->dS485()->getStatisticUpstream(
                                    dsm->getDSID(), 1, &length, data + index);
        index += length;
    }
    catch (Exception& ex)
    {
        FILE_LOG(logERROR) << "[Task Meter Statistic] Query from dsm " <<
                util::str(dsm->getDSID()) << " failed: " << ex.getMessage();
        return ex.getMessage();
    }

    uint16_t sversion = *(uint16_t *) data;
    boost::shared_ptr<StatisticMeter> pstats(new StatisticMeter(sversion));

    for (int i = 0; i < index; i++)
    {
        pstats->pushValue(data[i], 0);
    }
    pstats->parseValues();

    Database* db = Database::getInstance();
    db->saveStatisticMeter(0, dsm->getDSID(), pstats->m_value_map);

    DSID_PREFIX_LOG(logINFO, dsm->getDSID(), "stat_")
        << "Statistic Query Results:";
    boost::shared_ptr<std::map<statistic_field_t, std::string > > sresults =
        pstats->formatValues();
    for (std::map<statistic_field_t, std::string>::iterator it = sresults->begin();
         it != sresults->end();
         it++)
    {
        std::pair<std::string,std::string> desc = db->getStatisticDescription((*it).first);
        DSID_PREFIX_LOG(logINFO, dsm->getDSID(), "stat_") << desc.first << " = " << (*it).second;
        (*it).second = desc.first;
    }


    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    for (std::list<boost::shared_ptr<StatisticValue> >::iterator it = pstats->m_value_map->begin();
         it != pstats->m_value_map->end();
         it++)
    {
        json_object *vres = json_object_new_object();
        json_object_object_add(vres, "statid", json_object_new_int((*it)->m_field));
        json_object_object_add(vres, "value", json_object_new_int((*it)->m_value));
        json_object_object_add(res, sresults->at((*it)->m_field).c_str(), vres);
    }
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdSMWebHandler::resetStatistic()
{
    std::ostringstream result;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    dS_Assistant::getInstance()->dS485()->resetStatisticUpstream(
            dsm->getDSID());
    DSID_PREFIX_LOG(logINFO, dsm->getDSID(), "stat_") << "RESET";

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

void JSONdSMWebHandler::process()
{
    try
    {
        std::string result;
        std::string header = http_ok();

        if (m_command == "restartRegistration")
        {
            result = restartRegistration();
        }
        else if (m_command == "getCurrentTemperature")
        {
            result = getCurrentTemperature();
        }
        else if (m_command == "blink")
        {
            result = blink();
        }
        else if (m_command == "getStatistic")
        {
            result = getStatistic();
        }
        else if (m_command == "resetStatistic")
        {
            result = resetStatistic();
        }
        else
        {
            failure("unknown function called");
            return;
        }

        mg_write(m_connection, header.c_str(), header.length());
        mg_write(m_connection, result.c_str(), result.length());
    }
    catch (Exception ex)
    {
        failure(ex.getMessage().c_str());
    }
    catch (...)
    {
        failure("exception raised while processing call " + m_command);
    }
}

} // namespace

