/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __TASKS_H__
#define __TASKS_H__

#include <digitalSTROM/dsm-api-v2/dsm-api.h>
#include <boost/thread/mutex.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <boost/shared_ptr.hpp>
#include <string>
#include <time.h>
#include <queue>

#include "dsmapi-wrapper.h"
#include "hwdesc.h"
#include "hwstats.h"
#include "timer.h"

namespace dSA
{

class dS_Assistant;

// these values are saved in the database, so make sure to only add types,
// do not change or modify existing types
typedef enum
{
    UnknownTask             = 0,
    TestTransmissionQuality = 1,
    TestDevicePresence      = 2,
    TestDeviceStatistic     = 3,
    TestMeterStatistic      = 4,
    TestSystemMonitor       = 5,
    TestDeviceStatisticReset= 6,
    TestMeterStatisticReset = 7,
} task_type;

typedef enum
{
    UnknownStatus       = 0,
    Queued              = 1,
    Running             = 2,
    Finished            = 3,
    Cancelled           = 4,
    Sleeping            = 5,
    Paused              = 6
} task_status;

typedef enum
{
    Normal  = 0,
    High    = 1
} task_priority;

typedef enum
{
    kStateInit,
    kStateSetup,
    kStateStartup,
    kStateStartupWait,
    kStateCleanup,
    kStateCleanupWait,
    kStateVersion,
    kStateRequest,
    kStateResponse,
    kStateReady,
    kStateNext,
    kStateWait,
    kStateFinished
} task_state;

class dSTaskInfo
{
public:
    dSTaskInfo();
    void setDescription(std::string description);
    const std::string getName() const;
    const std::string getDescription() const;
    bool isValid() const;
    void invalidate();

    virtual ~dSTaskInfo();

protected:
    std::string m_name;
    std::string m_description;
    bool m_valid;
};

class dSLongTaskInfo : public dSTaskInfo
{
public:
    dSLongTaskInfo();
    void setTaskID(long long int task_id);
    long long int getTaskID() const;
    void setParentTaskID(long long int task_id);
    long long int getParentTaskID() const;
    unsigned char getProgress() const;
    std::string getStatusInfo() const;
    std::string getParameters() const;
    void setTaskStatus(task_status status);
    task_status getTaskStatus() const;
    std::string getTaskStatusText() const;
    virtual task_type getTaskType() const = 0;
    time_t getStartedTime() const;
    void setStartedTime(time_t t);
    time_t getFinishedTime() const;
    void setFinishedTime(time_t t);
    time_t getQueuedTime() const;
    void setQueuedTime(time_t t);
    bool isPausable() const;
    bool isPaused() const;
    void pause();
    void setPriority(task_priority priority);
    task_priority getPriority() const;


protected:
    long long int m_task_id;
    long long int m_parent_task_id;
    unsigned char m_progress;
    std::string m_status_info;
    task_status m_task_status;
    mutable boost::mutex m_status_mutex;
    std::string m_parameters;
    time_t m_started;
    time_t m_finished;
    time_t m_queued;
    bool m_paused;
    bool m_pausable;
    task_priority m_task_priority;
};

class dSTask : public dSTaskInfo
{
public:
    dSTask();
    virtual void run(dS_Assistant *) = 0;
};

class dSLongTask : public dSLongTaskInfo
{
public:
    dSLongTask();
    virtual void run(dS_Assistant *) = 0;
};

class TaskScheduler : public TimerSubscriber
{
public:
    TaskScheduler(boost::shared_ptr<dSTask> task, unsigned int delay);
    boost::shared_ptr<dSTask> getTask();
    virtual void notify();
private:
    boost::shared_ptr<dSTask> m_task;
};

class LongTaskScheduler : public TimerSubscriber
{
public:
    LongTaskScheduler(boost::shared_ptr<dSLongTask> task, unsigned int delay);
    boost::shared_ptr<dSLongTask> getTask();
    virtual void notify();
private:
    boost::shared_ptr<dSLongTask> m_task;
};

class GetBusMembersTask : public dSTask
{
public:
    GetBusMembersTask(bool async = true);
    virtual void run(dS_Assistant *);
private:
    bool m_async;
};

class GetDSMInfoTask : public dSTask
{
public:
    GetDSMInfoTask(dsid_t dsid, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    bool m_async;
};

class GetDSMZonesTask : public dSTask
{
public:
    GetDSMZonesTask(dsid_t dsid);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
};

class GetDSMZoneInfoTask : public dSTask
{
public:
    GetDSMZoneInfoTask(dsid_t dsid, uint8_t zone_index);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    uint8_t m_zone_index;
};

class ScanDSMTask : public dSTask
{
public:
    ScanDSMTask(dsid_t dsid, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    bool m_async;
};

class StartupDSMTask : public dSTask
{
public:
    StartupDSMTask(dsid_t dsid, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    bool m_async;
};

class SetDSMApartmentIdTask : public dSTask
{
public:
    SetDSMApartmentIdTask(dsid_t dsid, int apartment_id, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    int m_apartment_id;
    bool m_async;
};

class CheckDSMTask : public dSTask
{
public:
    CheckDSMTask(dsid_t dsid, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    bool m_async;
};

class CheckDSMThread : public dSLongTask
{
public:
    CheckDSMThread(int interval);
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
private:
    int m_interval;
    int m_state;
    int m_index;
    boost::shared_ptr<std::list<dsid_t> > m_dsms;
};

class GetAllDevicesTask : public dSTask
{
public:
    GetAllDevicesTask(dsid_t dsid);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
};

class GetActiveDevicesInZoneTask : public dSTask
{
public:
    GetActiveDevicesInZoneTask(dsid_t dsid, uint16_t zone_id);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    uint16_t m_zone_id;
};

class GetDeviceInfoTask : public dSTask
{
public:
    GetDeviceInfoTask(dsid_t dsm_dsid, uint16_t device_id);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    uint16_t m_device_id;

};

class TestTranmissionQualityTask :
    public dSLongTask,
    public dSMApiSelectiveListener,
    public boost::enable_shared_from_this< TestTranmissionQualityTask >
{
public:
    TestTranmissionQualityTask(int checksPerDevice, dsid_t dsm_dsid);
    virtual ~TestTranmissionQualityTask();
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
    virtual void callback(boost::shared_ptr<CBData> callback_data);
private:
    int m_checksPerDevice;
    dsid_t m_dsm_dsid;

    uint8_t m_result_down;
    uint16_t m_result_up;
    int m_result_err;
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > > m_dev_list;
    std::list<boost::shared_ptr<Device> >::const_iterator m_dev_current;
    size_t m_dev_count;
    size_t m_dev_total;
    int m_check_loop;
    task_state m_state;
};

class TestDevicePresenceTask :
    public dSLongTask,
    public dSMApiSelectiveListener,
    public boost::enable_shared_from_this< TestDevicePresenceTask >
{
public:
    TestDevicePresenceTask(dsid_t dsm_dsid, int interval_time);
    virtual ~TestDevicePresenceTask();
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
    virtual void callback(boost::shared_ptr<CBData> callback_data);

    typedef enum
    {
        pd_state_unknown = 1,
        pd_state_scheduled,
        pd_state_sent,
        pd_state_replied,
        pd_state_removed,
        pd_state_discovered
    } ping_state_t;

    class DeviceP : public Device
    {
    public:
        DeviceP(Device* parent);
        void setPingState(ping_state_t state) { m_ping_state = state; }
        ping_state_t getPingState() const     { return m_ping_state; };
    private:
        ping_state_t m_ping_state;
        uint8_t m_count_ping_sent;
        uint8_t m_count_ping_received;
    };

    typedef enum
    {
        pm_state_unknown = 1,
        pm_state_waiting,
        pm_state_ping_sent,
        pm_state_reply_received,
        pm_state_ready
    } ping_meter_state_t;

    class MeterP : public dSM
    {
    public:
        MeterP(dSM* parent, TestDevicePresenceTask* const task);
        void resetPingState();
        void getPingStatus(uint16_t* dev_count, uint16_t* tx_count, uint16_t* rx_count);
        uint16_t getPingRetries()         { return m_count_ping_timeout; }
        void signalPingSent(uint16_t device_id);
        void signalPingReply(uint16_t device_id);
        void signalPingTimeout();

        bool pingFinished()               { return m_state == pm_state_ready; }
        ping_meter_state_t getState()     { return m_state; }
        void setState(ping_meter_state_t state) { m_state = state; }

        void setSelectionA(uint16_t data) { m_a_selection = data; }
        uint16_t getSelectionA() const    { return m_a_selection; };
        void setSelectionB(uint16_t data) { m_b_selection = data; };
        uint16_t getSelectionB() const    { return m_b_selection; };
        void setSelectionD(uint32_t data) { m_d_selection = data; };
        uint32_t getSelectionD() const    { return m_d_selection; };

    private:
        TestDevicePresenceTask* const m_task;
        ping_meter_state_t m_state;
        uint16_t m_count_ping_dut;
        uint16_t m_count_ping_sent;
        uint16_t m_count_ping_received;
        uint16_t m_count_ping_timeout;
        uint16_t m_deviceid_dut;
        uint16_t m_a_selection;
        uint16_t m_b_selection;
        uint32_t m_d_selection;
    };

    boost::shared_ptr<MeterP> getDSM(dsid_t dsmid);
    boost::shared_ptr<DeviceP> getDevice(dsid_t dsmid, uint16_t devindex);
    boost::shared_ptr<DeviceP> getDevice(dsid_t dsmid, dsid_t devid);

private:
    dsid_t m_dsm_dsid;
    int m_interval_time;

    boost::shared_ptr<std::list<boost::shared_ptr<DeviceP> > > m_devs;
    boost::shared_ptr<std::list<boost::shared_ptr<MeterP> > > m_dsms;
    boost::shared_ptr<MeterP> m_dsm_under_test;
    uint16_t m_dsm_dev_count;
    uint16_t m_total_dev_count;
    uint16_t m_total_tx;
    uint16_t m_total_rx;
    uint16_t m_tx_count;
    uint16_t m_tx_count_equal;
    uint16_t m_tx_dev_count;
    uint16_t m_rx_count;

    int m_state;
    pthread_mutex_t m_ping_lock;
    pthread_cond_t m_ping_status_change;
};

class TestDeviceStatisticTask :
    public dSLongTask,
    public dSMApiSelectiveListener,
    public boost::enable_shared_from_this< TestDeviceStatisticTask >
{
public:
    TestDeviceStatisticTask(dsid_t dsm_dsid, dsid_t dev_dsid);
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
    virtual void callback(boost::shared_ptr<CBData> callback_data);
private:
    dsid_t m_dsm_dsid;
    dsid_t m_dev_dsid;

    task_state m_state;
    int m_offset;
    int m_length;
    int m_retry;
    int m_result_err;
    uint16_t m_result_value;
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > > m_dev_list;
    std::list<boost::shared_ptr<Device> >::const_iterator m_dev_current;
    size_t m_dev_count;
    boost::shared_ptr<Device> m_device;
    boost::shared_ptr<StatisticDevice> m_pstats;
};

class TestDeviceStatisticResetTask :
    public dSLongTask,
    public boost::enable_shared_from_this< TestDeviceStatisticResetTask >
{
public:
    TestDeviceStatisticResetTask(dsid_t dsm_dsid, dsid_t dev_dsid);
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
private:
    dsid_t m_dsm_dsid;
    dsid_t m_dev_dsid;

    task_state m_state;
    int m_retry;
    int m_errorcode;
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > > m_dev_list;
    std::list<boost::shared_ptr<Device> >::const_iterator m_dev_current;
    size_t m_dev_count;
    boost::shared_ptr<Device> m_device;
};

class TestMeterStatisticTask : public dSLongTask
{
public:
    TestMeterStatisticTask(dsid_t dsm_dsid);
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
private:
    dsid_t m_dsm_dsid;
};

class TestMeterStatisticResetTask : public dSLongTask
{
public:
    TestMeterStatisticResetTask(dsid_t dsm_dsid);
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
private:
    dsid_t m_dsm_dsid;
};

class TestSystemMonitorTask :
    public dSLongTask,
    public boost::enable_shared_from_this< TestSystemMonitorTask >
{
public:
    TestSystemMonitorTask(int profile, int duration);
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
private:
    int m_profile;
    int m_duration;
    task_state m_state;

    class SystemTestMethod
    {
    public:
        SystemTestMethod(task_type type, dsid_t target_dsm, dsid_t target_device, time_t dt);
        dSLongTask* getTask();
        std::string getName();
        bool isRunning();
        bool isToBeScheduled();
        long long int schedule(long long int parent_id);
    private:
        task_type m_type;
        boost::shared_ptr<dSLongTask> m_task;
        dsid_t m_dsm_dsid;
        dsid_t m_device_dsid;
    public:
        time_t m_deltatime;
        time_t m_lasttime;
    };

    boost::shared_ptr<std::list<boost::shared_ptr<SystemTestMethod> > >
        m_startup_methods, m_test_methods, m_cleanup_methods;
};

class BasicMonitorTask :
    public dSTask,
    public dSMApiSelectiveListener,
    public boost::enable_shared_from_this< BasicMonitorTask >
{
public:
    BasicMonitorTask();
    virtual void run(dS_Assistant *);
    virtual void callback(boost::shared_ptr<CBData> callback_data);

private:
    task_state m_state;
    std::queue<boost::shared_ptr<CBData> > m_queue;
    mutable boost::mutex m_mutex;
    bool m_db_storage;

    void process();
};

#define DB_CLEANUP_DAYS(d) (d*24*60*60)

class DatabaseCleanupTask : public dSTask
{
public:
    DatabaseCleanupTask(size_t max_db_filesize);
    virtual void run(dS_Assistant *);
private:
    size_t m_max_filesize;
    int m_state;
    int m_iteration;
};

}

#endif//__TASKS_H__
